/*
 * Created on Feb 24, 2005
 *
 */
package edu.buffalo.nsf.xmlcqa.data.dtd;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import edu.buffalo.nsf.xmlcqa.FatalError;
import edu.buffalo.nsf.xmlcqa.data.query.Query;
import edu.buffalo.nsf.xmlcqa.data.stream.Token;
import edu.buffalo.nsf.xmlcqa.tree.NodeCollector;
import edu.buffalo.nsf.xmlcqa.tree.OperationManager;
import edu.buffalo.nsf.xmlcqa.tree.TreeNode;

/**
 * A class representing Document Type Definition. DTD expresses the structure of
 * an XML documents using regular expression. For every a tag the method
 * {@link #getDefinition(Tag tag)} returns the regular expression which the
 * labels of children of the given tag should conform to. The {@link #init()}
 * method is used to initialize memoization in every regular expression that is
 * stored by the object of this class. The method
 * {@link #getMinTreeSize(Tag tag)} returns the size of a minimal XML document
 * whose root is labeled with <code>tag</code>. The values returned by this
 * method are computed once, during creation of the object, and then stored in
 * an array. This array can be accessed by method {@link #getMST()}. It is
 * assumed that when using this array, its values will not be changed. Any
 * violation of this contract may result in an undefined behaviour. The method
 * {@link #getMT(TreeNode parent, Tag tag, Query path, OperationManager mgr)}
 * returns the set of tree facts rooted at syntetic child of <code>parent</code>
 * w.r.t. the query <code>path</code> that are present in every minimal tree
 * conforming the DTD, which root is labeled with <code>tag</code>. For
 * efficiency resons, we memoize the result of previous calls of this method.
 * The freshness of the memoized values is established by pointer-wise comparing
 * of the attributes <code>path</code> and <code>mgr</code>.
 * 
 * @author staworko
 */
public class DTD {

	protected List defs;

	protected int[] mst;

	/*
	 * stores the maximum tree size that can be generated using the tags in the
	 * DTD
	 */
	protected int[] maxTree;

	protected Query path;

	protected OperationManager mgr;

	protected NodeCollector[] mt;

	protected boolean[] memoized;

	/**
	 * Standard constructor. The parameter <code>defsMap</code> is expected to
	 * map tags to regular expressions. By definition the label assigned to
	 * {@link Tag#PCDATA_TAG} is equal to {@link RegExpr#EPSILON}. During the
	 * construction of DTD the sizes of minimal trees are being computed for
	 * every tag and stored for further usage.
	 */
	public DTD(Map defsMap) {

		this.defs = new ArrayList(defsMap.size());

		for (int i = 0; i < Tag.getCount(); i++) {
			if (i == Tag.PCDATA_TAG.ID) {
				this.defs.add(Tag.PCDATA_TAG.ID, RegExpr.EPSILON);
			} else {
				this.defs.add(i, defsMap.get(Tag.getTag(i)));
			}

		}

		this.mst = findMST();
		this.maxTree = findMaxTree();
	}

	/**
	 * A method initializing the given object. This method forces all the
	 * preprocessing associated with construction of transition tables for the
	 * regluar expressions. No execution of this method is necessary for a
	 * proper functioning of DTD. However, this method allows to separate the
	 * cost of DTD preprocesing from other time costs of the system.
	 */
	public DTD init() {
		for (int i = 0; i < Tag.getCount(); i++) {
			getDefinition(Tag.getTag(i)).getStateCount();
			getDefinition(Tag.getTag(i)).getDelta();
			getDefinition(Tag.getTag(i)).getEpsilonDelta();
			getDefinition(Tag.getTag(i)).getMinCostDelta(this);
			getDefinition(Tag.getTag(i)).getFstTagTrDelta(this);
			getDefinition(Tag.getTag(i)).getFstIndTrDelta(this);
		}
		return this;
	}

	protected int[] findMST() {
		int[] res = new int[Tag.getCount()];
		boolean[] visited = new boolean[Tag.getCount()];
		boolean[] found = new boolean[Tag.getCount()];
		for (int i = 0; i < res.length; i++) {
			if (!found[i]) {
				res[i] = findMST(Tag.getTag(i), res, visited, found);
			}
		}
		return res;
	}

	protected int findMST(NDFA e, int[] res, boolean[] visited, boolean[] found) {
		if (e == RegExpr.EPSILON) {
			return 0;
		} else if (e instanceof Clousure) {
			NDFA e2 = ((Clousure) e).getExpr();
			return findMST(e2, res, visited, found);
		} else if (e instanceof Tag) {
			Tag t = (Tag) e;
			int i = t.ID;
			if (found[i]) {
				return res[i];
			} else if (visited[i]) {
				return Integer.MAX_VALUE;
			} else {
				visited[i] = true;
				res[i] = findMST(getDefinition(t), res, visited, found) + 1;
				visited[i] = false;
				return res[i];
			}
		} else if (e instanceof Concat) {
			Concat c = (Concat) e;
			return findMST(c.getExpr1(), res, visited, found)
					+ findMST(c.getExpr2(), res, visited, found);
		} else if (e instanceof Union) {
			Union u = (Union) e;
			int m1 = findMST(u.getExpr1(), res, visited, found);
			int m2 = findMST(u.getExpr2(), res, visited, found);
			return (m1 < m2) ? (m1) : (m2);
		} else {
			throw new FatalError("UNKN REG-EXPR type");
		}
	}

	protected int[] findMaxTree() {
		int[] res = new int[Tag.getCount()];
		boolean[] visited = new boolean[Tag.getCount()];
		boolean[] found = new boolean[Tag.getCount()];
		for (int i = 0; i < res.length; i++) {
			if (!found[i]) {
				res[i] = findMaxTree(Tag.getTag(i), res, visited, found);
			}
		}

		return res;
	}

	protected int findMaxTree(NDFA e, int[] res, boolean[] visited,
			boolean[] found) {

		if (e == RegExpr.EPSILON) {
			return 0;
		} else if (e instanceof Clousure) {
			return (Integer.MAX_VALUE);
		} else if (e instanceof Tag) {
			Tag t = (Tag) e;
			int i = t.ID;
			if (found[i]) {
				return res[i];
			} else if (visited[i]) {
				return Integer.MAX_VALUE;
			} else {
				visited[i] = true;
				res[i] = findMaxTree(getDefinition(t), res, visited, found) + 1;
				visited[i] = false;
				return res[i];
			}
		} else if (e instanceof Concat) {
			Concat c = (Concat) e;
			return findMaxTree(c.getExpr1(), res, visited, found)
					+ findMaxTree(c.getExpr2(), res, visited, found);
		} else if (e instanceof Union) {
			Union u = (Union) e;
			int m1 = findMaxTree(u.getExpr1(), res, visited, found);
			int m2 = findMaxTree(u.getExpr2(), res, visited, found);
			return (m1 > m2) ? (m1) : (m2);
		} else {
			throw new FatalError("UNKN REG-EXPR type");
		}
	}

	/**
	 * A standard class indicating improper DTD definition. DTD is proper if for
	 * every tag there exists a finite tree conforming the DTD with root labeled
	 * with the given tag. Detection of an improper DTD is performed only during
	 * computation of minimal sets of tree facts with the method
	 * {@link #getMT(TreeNode, Tag, Query, OperationManager)}
	 */
	public static class DTDCycleDetected extends Error {
		private static final long serialVersionUID = -2280167608356441161L;
	}

	/**
	 * Computes a minimal set of tree facts w.r.t. <code>path</code>, for
	 * trees whose root is labeled with <code>tag</code> rooted at a synthetic
	 * child of <code>parent</code>. The {@link NodeCollector} instance is
	 * created using supplied {@link OperationManager}.
	 */
	public NodeCollector getMT(TreeNode parent, Tag tag, Query path,
			OperationManager mgr) {
		if (this.path != path || this.mgr != mgr) {
			this.memoized = new boolean[Tag.getCount()];
			this.mt = new NodeCollector[Tag.getCount()];
		}

		if (!memoized[tag.ID]) {
			mt[tag.ID] = findMT(path, tag, mgr);
			memoized[tag.ID] = true;
		}

		TreeNode root = parent.getChild(-1, Token.createTag(tag));

		return mgr.substitute(mt[tag.ID], root);
	}

	private NodeCollector findMT(Query path, Tag t, OperationManager mgr) {
		boolean[] visited = new boolean[Tag.getCount()];
		Token tok = Token.createTag(t);
		NodeCollector root = mgr.createRoot(path, new TreeNode(tok));
		return findMT(t, root, 0, visited, mgr);
	}

	private NodeCollector findMT(NDFA e, NodeCollector parent, int i,
			boolean[] visited, OperationManager mgr) {
		if (e == RegExpr.EPSILON) {
			return parent;
		} else if (e instanceof Clousure) {
			NDFA e2 = ((Clousure) e).getExpr();
			return findMT(e2, parent, i, visited, mgr);
		} else if (e instanceof Tag) {
			Tag t = (Tag) e;
			int id = t.ID;
			if (t == Tag.PCDATA_TAG) {
				return parent;
			}
			if (visited[id]) {
				throw new DTDCycleDetected();
			} else {
				visited[i] = true;
				try {
					Token tok = Token.createTag(t.getName());
					NodeCollector child = mgr.createChild(parent, i++, tok);
					child = findMT(getDefinition(t), child, 0, visited, mgr);
					parent = mgr.hook(parent, child);
				} catch (DTDCycleDetected cd) {
					visited[i] = false;
					throw (DTDCycleDetected) cd.fillInStackTrace();
				}
				visited[i] = false;
				return parent;
			}
		} else if (e instanceof Concat) {
			Concat c = (Concat) e;
			parent = findMT(c.getExpr1(), parent, i, visited, mgr);
			return findMT(c.getExpr2(), parent, i, visited, mgr);
		} else if (e instanceof Union) {
			Union u = (Union) e;
			NodeCollector[] nc = new NodeCollector[2];

			try {
				nc[0] = findMT(u.getExpr1(), parent, i, visited, mgr);
			} catch (DTDCycleDetected cd1) {
			}
			try {
				nc[1] = findMT(u.getExpr2(), parent, i, visited, mgr);
			} catch (DTDCycleDetected dc2) {
			}

			if (nc[0] == null && nc[1] == null) {
				throw new DTDCycleDetected();
			}

			if (nc[0] == null) {
				return nc[1];
			} else if (nc[1] == null) {
				return nc[0];
			} else {
				return mgr.cut(nc);
			}
		} else {
			throw new FatalError("UNKN REG-EXPR type");
		}
	}

	/**
	 * Returs specification of children lables of a node labeled with
	 * <code>tag</code>.
	 */
	public NDFA getDefinition(Tag tag) {
		return (NDFA) defs.get(tag.ID);
	}

	/**
	 * Return the size of a minimal tree satisfing the DTD whose root node is
	 * labeled with <code>tag</code>.
	 */
	public int getMinTreeSize(Tag tag) {
		return mst[tag.ID];
	}
	
	/**
	 * Return the size of a maximal tree satisfing the DTD whose root node is
	 * labeled with <code>tag</code>.
	 */
	public int getMaxTreeSize(Tag tag) {
		return maxTree[tag.ID];
	}

	/**
	 * Returs the whole array of values suplied by the method
	 * {@link #getMinTreeSize(Tag)}. The returned array is an internal array
	 * storing all the values. It should not be modified.
	 */
	public int[] getMST() {
		return mst;
	}
	
	/**
	 * Returs the whole array of values suplied by the method
	 * {@link #getMaxTreeSize(Tag)}. The returned array is an internal array
	 * storing all the values. It should not be modified.
	 */
	public int[] getMaxTree() {
		return maxTree;
	}

	public String toString() {
		String dtd = new String("DTD:\n");
		for (int i = 0; i < Tag.getCount(); i++) {
			Tag t = Tag.getTag(i);
			dtd = dtd + t.getName() + " -> "
					+ ((NDFA) (getDefinition(t))).toString() + "\n";
		}
		return (dtd);
	}
}