/*
 * Created on Feb 24, 2005
 *
 */
package edu.buffalo.nsf.xmlcqa.data.dtd;

import edu.buffalo.nsf.xmlcqa.Constants;

/**
 * A class representing an nondeterminisic finite state automata (NDFA). 
 * The number of states of the NDFA can be obtained by {@link #getStateCount()}. 
 * The states are identified by natural numbers starting with <code>0</code>. 
 * The method {@link #getFinalStates()} indicates the final states. The transition 
 * relation is dividad into two parts: epsilon transitions (obtained with 
 * {@link #getEpsilonDelta()}) and transitions facilitated with a tag 
 * (obtained wiht {@link #getDelta()}). Although we distinguish between 
 * epsilon and tag transitions, the transitios facilitated with tags should be 
 * closed under epsilon transitions (i.e. if there is a transition from a state 
 * <code>i</code> to a state <code>j</code> using tag <code>t</code> and there is 
 * an epsilon transition from the state <code>j</code> to a state <code>k</code>, 
 * then there should be a transition from <code>i</code> to <code>k</code> with 
 * the tag <code>t</code>)  
 * 
 * <p>
 * An NDFA can be viewed as a directed graph, with its vertices being states and 
 * a pair of states being connected with an edge labeled with a tag if there is a 
 * transition from the first state to the second when reading the given tag. 
 * This graph is called here the <i>transition graph</i> of the given NDFA.
 * </p>
 * 
 * <p>
 * In the context of a DTD, crossing an edge of the transition graph can be charged 
 * with a cost corresponding to the minimal size of a document satifying the DTD and 
 * whose root is labeled with the tag of the edge (see the method {@link DTD#getMinTreeSize(Tag)}). 
 * Implementation of this class should provide methos for finding all cheapest paths
 * between every pair of vertices by means of methods: {@link #getMinCostDelta(DTD)},
 * {@link #getFstIndTrDelta(DTD)}, and {@link #getFstTagTrDelta(DTD)}. <br/>
 * The first method {@link #getMinCostDelta(DTD)} returns a two dimensional array 
 * containing minimal costs of paths in the transition graph between every pair of
 * vertices. The array returned by {@link #getFstIndTrDelta(DTD)} allows to construct 
 * all minimal path between any two vertices. The method {@link #getFstTagTrDelta(DTD)} 
 * provides tags that the edges of the paths are labeled with. And so for a DTD <code>d</code> 
 * and two vertices <code>i</code> and <code>j</code>, the minimal cost of 
 * getting from <code>i</code> to <code>j</code> is  
 * <code>getMinCostDelta(d)[i][j]</code>.  The values 
 * <code>getFstIndTrDelta(d)[i][j]</code> and 
 * <code>getFstTagTrDelta(d)[i][j]</code> consist of eaqual in length 
 * vectors of state identifiers and tags respectively. Assuming that the 
 * length of those vectors is <code>n</code>, denote the elements of those 
 * two vectors by <code>k_1, ..., k_n</code> (the states) and 
 * <code>t_1, ..., t_n</code> (the tags). For any <code>m &lt;= n</code>, if 
 * <code>i,v_1,...v_s,k_m</code> is a ceapest path from <code>i</code> to <code>k_m</code>,
 * then <code>i,v_1,...,v_s,k_m,j</code> is a minimal path between from 
 * <code>i</code> to <code>j</code>, and the edge from <code>k_m</code> to 
 * <code>j</code> is labeled with <code>t_m</code>. This resoning captures all 
 * cheapest paths from <code>i</code> to <code>j</code>.    
 * </p>
 *    
 * @author staworko
 */
public abstract class NDFA implements Constants
{
    /**
     * Returns the number of states used by the underlying NDFA. 
     * The states are number with indexes 0 and higher (similar to arrays).
     */
    public abstract int           getStateCount();
    
    /**
     * Returns a boolean vector specifying final states. A state <code>i</code> is final 
     * if and only if <code>getFinalStates()[i]</code> is <code>true</code>.  
     */
    public abstract boolean[]     getFinalStates();
    
    /**
     * Returns the tag controled part of the transition relation. 
     * <code>getDelta()[t][i][j]</code> is <code>true</code> if
     * and only if there is transition with a tag of {@link Tag#ID identifier} 
     * equal to <code>t</code>, from the state <code>i</code> to the state <code>j</code>  
     *  
     */
    public abstract boolean[][][] getDelta();
    
    /**
     * Returns the epsilon part of the transition relation. <code>getEpsilonDelta()[i][j]</code>
     * is <code>true</code> if and only if there is an epsilon transition from the state 
     * <code>i</code> to the state <code>j</code>. 
     */
    public abstract boolean[][]   getEpsilonDelta();
    
    /**
     * Returns an array containing costs of cheapest paths 
     * between every pari of states. The costs of the edges 
     * depend on the values of {@link DTD#getMinTreeSize(Tag)}.
     */
    public abstract int[][] 	  getMinCostDelta(DTD d);
    
    /**
     * Returns a two dimensional array containing vectors of 
     * suffixes of the cheapest paths between each pair of vertices. 
     */
    public abstract int[][][] 	  getFstIndTrDelta(DTD d);
    
    /**
     * Returns a two dimentional array containing vectors of 
     * suffizes of labels of the cheapest paths between each 
     * pair of vertices.
     */
    public abstract Tag[][][] 	  getFstTagTrDelta(DTD d);

}