/*
 * Created on Jun 10, 2005
 *
  */
package edu.buffalo.nsf.xmlcqa.data.dtd;


/**
 * An abstract class facilitating construction of regular expressions. The semantics 
 * of regular expression is expressed by means of nondeterministic finite automata.  
 * Therefore this class extends {@link NDFA}. Also this class 
 * allows to delegate construction of all methods specified by 
 * {@link NDFA} by implementing four of it abstract methods: 
 * {@link #_getStateCount()}, 
 * {@link #_fillFinalStates(boolean[] f , int offset)}, 
 * {@link #_fillDelta(boolean[][][] delta, int offset)}, and
 * {@link #_fillEpsilonDelta(boolean[][] epsilon, int offset)}. 
 * Three of those methods have a parameter <code>offset</code>, which facilitates 
 * structural composition of regular expressions. An object implementing those methods
 * should assume that numbers of its states start with the value <code>offset</code>.    
 * 
 * <p>
 * The values specified by abstract methods in {@link NDFA} are memoized once they
 * have been computed. Should the value of the parameters change, the vlaues will be 
 * recomputed.  
 * </p>
 *   
 * @author staworko
 */
public abstract class RegExpr extends NDFA
{
    /**
     * Should return the number of states the implementation is using. This method will be 
     * executed only once and its result stored and returned with each call of 
     * {@link #getStateCount()} method (memoization). 
     */
    protected abstract int  _getStateCount();    
    
    /**
     * Used to indicate final states of a regular expression whose position is relative to 
     * <code>offset</code>.  
     */
    protected abstract void _fillFinalStates(boolean[] f, int offset);
    
    /**
     * Used to fill tag driven part of the transition table of a regular expression 
     * whose position if relative to <code>offset</code>. The implementing class doesn't 
     * have indicate transitions obtained in a consequence of clousure under epsilon transition.  
     * This clousure will be computed automatically by this class.     
     */
    protected abstract void _fillDelta(boolean[][][] delta, int offset);
    
    /**
     * Used to fill epsilon part of the transition talbe of a regular expressio 
     * whose possition is relative to <code>offset</code>. The implementing class doesn't 
     * have indicate transitions obtained in the consequence of the clousure. This clousure 
     * will be computed by this class.
     */
    protected abstract void _fillEpsilonDelta(boolean[][] delta, int offset);    

    /*
     * Memoization variables
     */
    private int           _stateCount    = -1;
    private boolean[][][] _delta         = null;
    private boolean[][]   _epsilonDelta  = null;
    private boolean[]     _finalStates   = null;

    /*
     * Relative memoization (w.r.t dtd) 
     */
    private DTD           _dtd           = null;
    private Tag[][][]     _fstTagTrDelta = null;
    private int[][][]     _fstIndTrDelta = null;
    private int[][]       _minCostDelta  = null;
    /**
     * A constant representing epsilon regular expression. 
     * This regular expression defines set consisting of only one word: the empty word.   
     */
    public static final RegExpr EPSILON      = new RegExpr() {
    
                                            public String toString()
                                            {
                                                return "EMPTY";
                                            }
    
                                            protected int _getStateCount()
                                            {
                                                return 1;
                                            }
    
                                            protected void _fillDelta(
                                                                      boolean[][][] delta,
                                                                      int offset)
                                            {
                                            }
    
                                            protected void _fillEpsilonDelta(
                                                                             boolean[][] delta,
                                                                             int offset)
                                            {
                                            }
    
                                            protected void _fillFinalStates(boolean[] f, int offset)
                                            {
                                                f[offset] = true;
                                            }
    
                                        };

    public int getStateCount()
    {
        if (_stateCount == -1) {
            _stateCount = _getStateCount();
        }
        
        return _stateCount;
    }

    public boolean[][][] getDelta()
    {
        if (_delta == null) {
            _delta = _getDelta();
        }
        
        return _delta;
    }

    public boolean[][] getEpsilonDelta()
    {
        if (_epsilonDelta == null) {
            _epsilonDelta = _getEpsilonDelta();
        }
        
        return _epsilonDelta;
    }
    
    public boolean[] getFinalStates()
    {
        if(_finalStates == null) {
            _finalStates = _getFinalStates();
        }
        
        return _finalStates; 
    }

    public int[][] getMinCostDelta(DTD d)
    {
        checkMemoization(d);
        return _minCostDelta;
    }
    
    public int[][][] getFstIndTrDelta(DTD d)
    {
        checkMemoization(d);
        return _fstIndTrDelta;
    }

    public Tag[][][] getFstTagTrDelta(DTD d)
    {
        checkMemoization(d);
        return _fstTagTrDelta;
    }


    private boolean[][] _getEpsilonDelta()
    {
        boolean[][] delta = new boolean[getStateCount()][];
        for (int i = 0; i < delta.length; i++) {
            delta[i] = new boolean[getStateCount()];
            for (int j = 0; j < delta[i].length; j++) {
                delta[i][j] = i == j;
            }
        }
        _fillEpsilonDelta(delta, 0);
        closure(delta);
        return delta;
    }

    private boolean[][][] _getDelta()
    {
        boolean[][][] delta = new boolean[Tag.getCount()][][];
        for (int i = 0; i < delta.length; i++) {
            delta[i] = new boolean[getStateCount()][];
            for (int j = 0; j < delta[i].length; j++) {
                delta[i][j] = new boolean[getStateCount()];
                for (int k = 0; k < delta[i][j].length; k++) {
                    delta[i][j][k] = false;
                }
            }
        }
        _fillDelta(delta, 0);
        
        boolean[][] epsilonDelta = getEpsilonDelta();
        
        for (int i = 0; i < Tag.getCount(); i++) {
            multiply(delta[i],epsilonDelta,delta[i]);
            multiply(epsilonDelta,delta[i],delta[i]);
        }
        
        return delta;
    }
    
    private boolean[] _getFinalStates()
    {
        boolean[] finalStates = new boolean[getStateCount()];

        _fillFinalStates(finalStates,0);
        
        return finalStates;
    }

    private void checkMemoization(DTD d)
    {
        if (d != _dtd) {
            _dtd = d;
            this._minCostDelta  = null;
            this._fstTagTrDelta = null;
            this._fstIndTrDelta = null;
            _updateTrDeltas(d);
        }
    }

    private void multiply(boolean[][] a, boolean[][] b, boolean[][] c)
    {
        int n = a.length;
        
        for (int i = 0; i < n; i++) {
            for (int j = 0; j < n; j++) {
                for (int k = 0; k < n; k++) {
                    c[i][j] = c[i][j] || (a[i][k] && b[k][j]);
                }
            }
        }
    }
    
    private void closure(boolean[][] a)
    {
        int n = a.length;
        int log_n = (int) Math.ceil(Math.log((double)n));
        for (int i = 0; i < log_n; i++) {
            multiply(a,a,a);
        }
    }

    
    private int[] set(int x)
    {
        int[] xs = new int[1];
        xs[0] = x;
        
        return xs;
    }
    
    private Tag[] set(Tag x)
    {
        Tag[] xs = new Tag[1];
        xs[0] = x;
        
        return xs;
    }
    

    private int[] add(int[] xs, int x)
    {
        int[] ys = new int[xs.length + 1];
        
        for (int i = 0; i < xs.length; i++){
            ys[i] = xs[i]; 
        }
        
        ys[xs.length] = x;
        
        return ys;
    }
    
    private Tag[] add(Tag[] xs, Tag x)
    {
        Tag[] ys = new Tag[xs.length + 1];
        
        for (int i = 0; i < xs.length; i++){
            ys[i] = xs[i]; 
        }
        
        ys[xs.length] = x;
        
        return ys;
    }
    
    private boolean contains (Tag[] ts, int[] xs, Tag t, int x)
    {
        for (int i = 0; i < ts.length; i++) {
            if (ts[i] == t && xs[i] == x){ 
                return true; 
            }
        }
        
        return false;
    }

    
    private void _updateTrDeltas(DTD d) 
    {
        boolean[][][] delta        = getDelta();
        boolean[][]   epsilonDelta = getEpsilonDelta();
        
        int[] mst = d.getMST();
        int n = getStateCount();
        
        int[][]   mcd = new int[n][];
        Tag[][][] ftd = new Tag[n][][];
        int[][][] fid = new int[n][][];
        
        for (int i = 0; i < n; i++) {
            mcd[i] = new int[n];
            ftd[i] = new Tag[n][];
            fid[i] = new int[n][];
            for (int j = 0; j < n; j++) {
                mcd[i][j] = (epsilonDelta[i][j]) ? (0) : (INF);
                ftd[i][j] = new Tag[0];
                fid[i][j] = new int[0];
                
                for (int t = 0; t < Tag.getCount(); t++) {
                    if (delta[t][i][j]) {
                        if (mcd[i][j] == mst[t]) {
                            if (contains(ftd[i][j],fid[i][j],Tag.getTag(t),j)) {  
                                continue;
                            }
                            ftd[i][j] = add(ftd[i][j],Tag.getTag(t));
                            fid[i][j] = add(fid[i][j],i);
                        } else if (mcd[i][j] > mst[t]) {
                            mcd[i][j] = mst[t];
                            ftd[i][j] = set(Tag.getTag(t));
                            fid[i][j] = set(i);                            
                        }                        
                    }
                }
            }
        }
        
        for (int s = 1; s < n ; s++) {
            for (int i = 0; i < n; i++) {
                for (int j = 0; j < n; j++) {
                    for (int k = 0; k < n; k++) {
                        for (int t = 0; t < Tag.getCount(); t++) {
                            if (! epsilonDelta[i][k] && delta[t][k][j]) {
                                int e = mcd[i][k] + mst[t];
                                if (mcd[i][j] == e) {
                                    if (contains(ftd[i][j],fid[i][j],Tag.getTag(t),k)) {  
                                        continue;
                                    }
                                    ftd[i][j] = add(ftd[i][j],Tag.getTag(t));
                                    fid[i][j] = add(fid[i][j],k);
                                } else if (mcd[i][j] > e) {
                                    mcd[i][j] = e;
                                    ftd[i][j] = set(Tag.getTag(t));
                                    fid[i][j] = set(k);                            
                                }    
                            }
                        }
                    }
                }
            }
            
        }
        
        _minCostDelta = mcd;
        _fstTagTrDelta = ftd;
        _fstIndTrDelta = fid;
        
    }
}
