/*
 * Created on May 12, 2005
 *
 */
package edu.buffalo.nsf.xmlcqa.data.dtd;

import edu.buffalo.nsf.xmlcqa.data.stream.Token;
import edu.buffalo.nsf.xmlcqa.data.stream.TokenStream;

/**
 * Automaton checking the conformance of a 
 * {@link edu.buffalo.nsf.xmlcqa.data.stream.TokenStream token stream}
 * with a given {@link DTD}. The whole functionality of this class is 
 * is implemented with static methods. The main method is 
 * {@link #validate(TokenStream, DTD)} which returns the number of detected 
 * errors (times the sequences of labels of children of a given node do not 
 * conform to the regular expression). 
 * 
 * @author staworko
 *  
 */
public class ValidationAutomaton
{

    /**
     * Calculates the number of errors in <code>ts</code> w.r.t. <code>d</code>. 
     * Result equal to <code>0</code> indicates stream conforming to the DTD. 
     */
    public static int validate(TokenStream ts, DTD d)
    {
        return validate(ts, d, ts.next().getTag());
    }

    private static void reportError(String msg)
    {
        System.out.println("Error in the stream: " + msg);
    }

    private static boolean[] clearState(boolean[] q)
    {
        for (int i = 0; i < q.length; i++) {
            q[i] = false;
        }
        return q;
    }

    private static boolean[] startState(NDFA def)
    {
        boolean[] start = new boolean[def.getStateCount()];
        start[0] = true;
        multiply(start, def.getEpsilonDelta(), start);
        return start;
    }

    private static int validate(TokenStream ts, DTD d, Tag root)
    {
        NDFA def = d.getDefinition(root);
        if (def == RegExpr.EPSILON) {
            return validateEmpty(ts,d,root);
        }
        boolean[] finalStates = def.getFinalStates();        
        boolean[] curr = startState(def);
        boolean[] next = new boolean[def.getStateCount()];
        boolean[] temp = null;
        int errors = 0;
        Token t = null;
        
        
        
        while (ts.hasNext() && (t = ts.next()) != Token.CLOSING_TOKEN) {
            if (t.isOpening()) {
                errors += validate(ts, d, t.getTag());
            }
            consume(def.getDelta(), t.getTag(), curr, next);
            temp = curr;
            curr = next;
            next = clearState(temp);
        }
        
        if (! scalarMultiply(curr,finalStates)) {
            errors++;
        }
        
        return errors;
    }

    private static int validateEmpty(TokenStream ts, DTD d, Tag root)
    {
        Token t = null;
        int paren = 1;
        int errors = 0;
        
        
        do {
            t = ts.next();
            switch(t.getType()) {
            case Token.CLOSING:
                paren--;
                break;
            case Token.OPENING:
                if (errors == 0) {
                    errors = 1;
                }
                paren++;
                break;
            case Token.TEXT:
                if (errors == 0) {
                    errors = 1;
                }
                break;
            }
        } while (paren > 0);
        
        return errors;
    }

    private static void consume(boolean[][][] delta, Tag t, boolean[] curr,
                                  boolean[] next)
    {
        multiply(curr, delta[t.ID], next);
    }

    /**
     * a times b gives c.
     */
    private static void multiply(boolean[] a, boolean[][] b, boolean[] c)
    {
        for (int i = 0; i < c.length; i++) {
            for (int j = 0; j < a.length && !c[i]; j++) {
                c[i] = a[j] && b[j][i];
            }
        }
    }
    
    private static boolean scalarMultiply(boolean[] a, boolean[] b) {
        for (int i = 0; i < a.length; i++) {
            if (a[i] && b[i]) {
                return true;
            }
        }
        
        return false;
    }
}