/*
 * Created on Feb 24, 2005
 *
 */
package edu.buffalo.nsf.xmlcqa.data.stream;

import java.util.Arrays;
import java.util.List;
import edu.buffalo.nsf.xmlcqa.Constants;

/**
 * @author staworko
 * 
 *  
 */
public class StreamDist implements Constants
{

    protected static final int HORIZ = -1;
    protected static final int DIAG  = 0;
    protected static final int VERT  = 1;

    protected static class Path
    {

        protected static final Path EMPTY = new Path(null, 0, 0);
        protected static final Path ROOT  = new Path(EMPTY, 0, 0);
        protected Path              parent;
        protected int               dir;
        protected int               cost;

        protected Path(Path parent, int dir, int cost)
        {
            this.parent = parent;
            this.dir = dir;
            this.cost = cost;
        }

        public Path consume(Token t, int dir)
        {
            if (this == EMPTY) {
                throw new IllegalArgumentException("Path is empty.");
            }
            switch (t.getType()) {
            case Token.OPENING:
                return new Path(this, dir, 1);
            case Token.CLOSING:
                return parent;
            case Token.TEXT:
            default:
                return this;
            }
        }

        public Path consume(Token x, Token y)
        {
            if (x.getType() != y.getType()) {
                throw new IllegalArgumentException(
                                                   "Incompatible types of tokens.");
            }
            switch (x.getType()) {
            case Token.TEXT:
                return this;
            case Token.OPENING:
                return new Path(this, DIAG, (x.equals(y)) ? (0) : (1));
            case Token.CLOSING:
            default:
                return parent;
            }
        }

        public int getDir()
        {
            return dir;
        }

        public int getCost()
        {
            return cost;
        }

        public String toString()
        {
            if (this == ROOT) {
                return "root";
            } else if (this == EMPTY) {
                return "empty";
            } else {
                switch (dir) {
                case VERT:
                    return "" + parent + "|";
                case HORIZ:
                    return "" + parent + "-";
                case DIAG:
                    return "" + parent + "\\";
                }
                return "ERROR";
            }
        }
    }

    protected static class Cell
    {

        public Path path; // path
        public int  cost; // cost

        public static Cell createRoot()
        {
            return new Cell(Path.ROOT, 0);
        }

        public static Cell create()
        {
            return new Cell(Path.EMPTY, 0);
        }

        protected Cell(Path path, int cost)
        {
            this.path = path;
            this.cost = cost;
        }

        public boolean isEmpty()
        {
            return path == Path.EMPTY;
        }

        /**
         * calculates cost after consumption.
         * 
         * @return INF if operation is inappropriate
         */
        public int consumption(Token x, int dir)
        {
            if (isEmpty()) {
                return INF;
            } else if (path.getDir() == DIAG && x.isClosing()) {
                return INF;
            } else if (path.getDir() == -1 * dir) {
                return INF;
            }
            return cost + 1;
        }

        /**
         * calculates cost after consumption.
         * 
         * @return INF if operation is inappropriate
         */
        public int consumption(Token x, Token y)
        {
            if (isEmpty()) {
                return INF;
            }
            if (x.getType() != y.getType()) {
                return INF;
            } else if (path.getDir() != DIAG) {
                return INF;
            }
            if (x.isClosing()) {
                return cost + path.getCost();
            } else {
                return cost + ((x.equals(y)) ? (0) : (1));
            }
        }

        public void consume(Cell c, Token x, int dir)
        {
            path = c.path.consume(x, dir);
            cost = c.consumption(x, dir);
        }

        public void consume(Cell c, Token x, Token y)
        {
            path = c.path.consume(x, y);
            cost = c.consumption(x, y);
        }

        public void setEmpty()
        {
            path = Path.EMPTY;
            cost = 0;
        }

        public int getCost()
        {
            return cost;
        }

        public void copy(Cell c)
        {
            this.path = c.path;
            this.cost = c.cost;
        }

        public String toString()
        {
            return "" + path + ":" + cost;
        }
    }

    protected boolean modify;
    protected int     h;
    protected int     size;
    protected Cell[]  col;
    protected Cell[]  ncol;
    private Token[]   y;
    protected int     i;
    protected Cell[]  row;
    protected Cell[]  nrow;
    private Token[]   x;
    protected int     j;

    public StreamDist(int h, boolean modify)
    {
        this.modify = modify;
        this.h = h;
        this.size = 2 * h + 1;
        this.y = new Token[size];
        this.row = new Cell[size];
        this.nrow = new Cell[size];
        this.x = new Token[size];
        this.col = new Cell[size];
        this.ncol = new Cell[size];
        for (int k = 0; k < size; k++) {
            row[k] = Cell.create();
            nrow[k] = Cell.create();
            col[k] = Cell.create();
            ncol[k] = Cell.create();
        }
        row[0] = col[0] = Cell.createRoot();
    }

    protected void swapCol()
    {
        Cell[] tmp = col;
        col = ncol;
        ncol = tmp;
    }

    protected void swapRow()
    {
        Cell[] tmp = row;
        row = nrow;
        nrow = tmp;
    }

    protected void swap(int orient)
    {
        if (orient == VERT) {
            swapCol();
        } else {
            swapRow();
        }
    }

    public int find(TokenStream xs, TokenStream ys)
    {
        while (xs.hasNext() || ys.hasNext()) {
            if (xs.hasNext()) {
                x[++i % size] = xs.next();
                fillVector(x, i, y, j, col, ncol, row, VERT);
            }
            if (ys.hasNext()) {
                y[++j % size] = ys.next();
                fillVector(y, j, x, i, row, nrow, col, HORIZ);
            }
        }
        return col[j % size].getCost() / 2;
    }

    protected void fillVector(Token[] x, int i, Token[] y, int j, Cell[] col,
                              Cell[] ncol, Cell[] row, int orient)
    {
        int b = (j - (size + 1) > 0) ? (j - (size + 1)) : (0);
        ncol[b % size].setEmpty();
        for (int k = b + 1; k <= j; k++) {
            int vert = ncol[(k - 1) % size].consumption(y[k % size], VERT
                    * orient);
            int diag = col[(k - 1) % size]
                    .consumption(x[i % size], y[k % size]);
            int horiz = col[k % size].consumption(x[i % size], HORIZ * orient);
            if (vert == INF && diag == INF && horiz == INF) {
                ncol[k % size].setEmpty();
                continue;
            }
            if (vert <= diag && vert <= horiz) {
                ncol[k % size].consume(ncol[(k - 1) % size], y[k % size], VERT
                        * orient);
            } else if (diag <= vert && diag <= horiz) {
                ncol[k % size].consume(col[(k - 1) % size], x[i % size], y[k
                        % size]);
            } else {
                ncol[k % size].consume(col[k % size], x[i % size], HORIZ
                        * orient);
            }
        }
        row[i % size].copy(ncol[j % size]);
        swap(orient);
    }

    protected static String list2str(List l, int i, int size)
    {
        String res = "[";
        int start = (i > size) ? (i - size) : (0);
        for (int k = start; k < i; k++) {
            res += "" + l.get(k % size);
            if (k < i - 1) {
                res += ",";
            }
        }
        return res + "]";
    }

    protected static String arr2str(Token[] t, int i, int size)
    {
        return list2str(Arrays.asList(t), i, size);
    }

    protected static String arr2str(Cell[] c, int i, int size)
    {
        return list2str(Arrays.asList(c), i, size);
    }

    public String toString()
    {
        return "X       :" + arr2str(x, i + 1, size) + "\n" + "NCol    :"
                + arr2str(ncol, j + 1, size) + "\n" + "Col(i=" + (i) + "):"
                + arr2str(col, j + 1, size) + "\n" + "Y       :"
                + arr2str(y, j + 1, size) + "\n" + "Row     :"
                + arr2str(nrow, i + 1, size) + "\n" + "Row(j=" + (j) + "):"
                + arr2str(row, i + 1, size) + "\n";
    }
}