/*
 * Created on Feb 23, 2005
 *
 * TODO To change the template for this generated file go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
package edu.buffalo.nsf.xmlcqa.data.stream;

import java.util.ArrayList;
import java.util.List;
import edu.buffalo.nsf.xmlcqa.data.dtd.Tag;

/**
 * Basic unit of XML documents. Tokens are used to convey the structure 
 * and data in the XML document in a streamed fashion. 
 * There are three types of tokens 
 * (method {@link #getType()}): 
 * <ul>
 * <li> {@link #OPENING} for tokens which denote a node in the XML document 
 * labeled with a tag ({@link #getTag()}). Every opening token should be balanced
 * with a closing token: any information between a mathching pair of opening and closing 
 * tokens contributes to the children of the node described by the opening token. 
 * This type of a token can be created with methods:
 * {@link #createTag(int tagID)}, {@link #createTag(Tag tag)}, or 
 * {@link #createTag(String tagName)}.    
 * </li> 
 * <li> {@link #TEXT} for tokens which describe a text node. The textual 
 * data can be accessed with the method {@link #getText()}. Becuase a text node
 * does not have any children, this node is not balanced with a 
 * closing token. However, similarly to an opening token, this token has also tag 
 * assigned which is equal to {@link Tag#PCDATA_TAG}. This type of a token can be 
 * created with method {@link #createText(String textValue)}
 * </li>
 * <li> {@link #CLOSING} for tokens balancing the opening tokens. For the sake 
 * of simplicity a closing token does not carry any information on the token 
 * it balances (the method {@link #getTag()} returns <code>null</code>). This allows 
 * us to create a constant {@link #CLOSING_TOKEN}. No other instances of a closing 
 * token should be created.
 * </li>
 * </ul>
 * <p>
 * Similarly to the class {@link Tag} this class memoizes all opening tokens. 
 * This behaviour prevents from multiple copies of the same object and allows to 
 * compare opening tokens by pointer comparizon. This <b>does not</b> apply to text 
 * tokens whose creation in not controled in any way, and it is posilbe that 
 * two text tokens with the came textual information are represented by 
 * two different instances of this class.
 * </p> 
 *   
 * @author staworko
 *  
 */
public class Token
{

    /**
     * Constant denoting opening token type.
     */
    public  static final int   OPENING       = 1;
    
    /**
     * Constant denoting closing token type.
     */
    public  static final int   CLOSING       = 2;
    
    /**
     * Constant denoting text token type.
     */
    public  static final int   TEXT          = 3;
    
    /**
     * Constant used to denote closing token.
     */
    public  static final Token CLOSING_TOKEN = new Token();
    
    private static final List  tokens        = new ArrayList();

    /**
     * Creates an opening token. This token is labeled 
     * with <code>tag</code>. 
     */
    public static Token createTag(Tag tag) 
    {
        if (tag == Tag.PCDATA_TAG) {
            throw new IllegalArgumentException("Wrong tag type (only non-PC_DATA tags allowed");
        }

        return createTag(tag.ID);
    }
    
    /**
     * Creates an opening token. This token is labeled with tag 
     * whose {@link Tag#ID} is equal to <code>tagID</code> 
     */
    public static Token createTag(int tagID)    
    {
        if (tagID == Tag.PCDATA_TAG.ID) {
            throw new IllegalArgumentException("Wrong tag ID (only non-PC_DATA tags allowed");
        }

        if (tokens.size() <= tagID) {
            for (int i = tokens.size(); i <= tagID; i++) {
                tokens.add(i,new Token(Tag.getTag(i)));
            }
        }
        
        return (Token) tokens.get(tagID);
    }
    
    
    /**
     * Creates an opening token. This token is labeled with a tag 
     * whose name is <code>tagName</code>
     */
    public static Token createTag(String tagName)
    {
        return createTag(Tag.get(tagName));
    }

    /**
     * Creates a text token. The textual data carried by this 
     * token is specified by <code>textValue</code>.
     */
    public static Token createText(String textValue)
    {
        return new Token(textValue);
    }

    private int  type;
    private Tag  tag;
    private String text;

    private Token()
    {
        this.type = CLOSING;
    }

    private Token(Tag tag)
    {
        this.type = OPENING;
        this.tag = tag;
    }

    private Token(String text)
    {
        this.type = TEXT;
        this.tag = Tag.PCDATA_TAG;
        this.text = text;
    }

    /**
     * Returns <code>true</code> if an only if this tag is opening.
     */
    public boolean isOpening()
    {
        return type == OPENING;
    }

    /**
     * Returns <code>true</code> if and only if the tag is closing.
     */
    public boolean isClosing()
    {
        return type == CLOSING;
    }

    /**
     * Returns <code>true</code> if and only if it's a text tag. 
     */
    public boolean isText()
    {
        return type == TEXT;
    }

    /**
     * Returns the tag. 
     * This method should be used only if this token is opening or textual. 
     */
    public Tag getTag()
    {
        return tag;
    }

    /**
     * Returns the textual information. This method should be used 
     * only if this token is textual. 
     */
    public String getText()
    {
        return text;
    }

    /**
     * Returns the type of the token.
     */
    public int getType()
    {
        return type;
    }

    /**
     * REturns the hash code value of this token.
     */
    public int hashCode()
    {
        switch (type) {
        case OPENING:
            return tag.ID;
        case TEXT:
            return text.hashCode();
        default:
            return 0;
        }
    }

    /**
     * Find if this token is equal to a given one.  
     */
    public boolean equals(Object o)    
    {
        if (this == o) {
            return true;
        } else if (o instanceof Token) {
            return equal(this, (Token) o);
        } else {
            return false;
        }
    }

    private static boolean equal(Token t1, Token t2)
    {
        if (t1.getType() == t2.getType()) {
            switch (t1.getType()) {
            case CLOSING:
                return true;
            case OPENING:
                return t1.getTag() == t2.getTag();
            case TEXT:
                return t1.getText().equals(t2.getText());
            }
        }
        return false;
    }

    /**
     * Returns a string representation of this token. 
     */
    public String toString()
    {
        if (isClosing()) {
            return ">";
        } else if (isText()) {
            return text.toString();
        } else if (isOpening()) {
            return "<" + tag.getName();
        } else {
            return "UNK-TKN";
        }
    }
}